#!/usr/bin/env node

var args = require('yargs'),
    M = require('mstring'),
    pkg = require('../package.json'),
    EnvController = require('../lib/controllers/EnvironmentController'),
    AppController = require('../lib/controllers/AppController'),
    prompt = require('prompt'),
    path = require('path');


var usage = M(function() {

    /***

     Commands:
       $0 new-app             <app-name>        [-t --title --description]
       $0 publish-app                           [-e -n -u -p --dir]
       $0 remove-app-remote   <app-name>        [-e -n -u -p]
       $0 list-apps                             [-e -n -u -p]
       $0 add-profile         <profile name>    [-e -n -u -p]
       $0 remove-profile      <profile name>    [-e -n -u -p]
       $0 list-profiles                         [-e -n -u -p]
     ***/

});

// TODO: make this better!
var argv = args
    .usage(usage)
    .options('t', {alias:'type', describe: 'The project type (custom | angular | app-generator).'})
    .options('title', {describe: 'The display name of the app.'})
    .options('description', {describe: 'The description of the app.'})
    .options('e', {alias:'environment', describe: 'The environment to use. Not needed if you want to use the default environment.'})
    .options('n', {alias:'namespace',   describe: 'The namespace. Not needed if environment is specified.'})
    .options('u', {alias:'user',    describe: 'The username. Not needed if environment is specified.'})
    .options('p', {alias:'password',    describe: "The password. Not needed if environment is specified."})
    .options('dir', {describe: 'The directory within the app folder that contains the app files to be published. This defaults to /www'})
    .argv;


//Just want the version
if(argv.version){
    console.log(pkg.version);
    return;
}

//Just want the help
if(argv.help){
    args.showHelp();
    return;
}

function AppTools(options) {

    this.cwd          = process.cwd();   // save this before loadProjectSync changes it
    this.helpers      = require('../lib/helpers')(argv);

    this.project      = this.helpers.loadProjectSync();
    this.createEnv    = this.helpers.createEnv;

    this.controllers  = {
        env: EnvController(this.project, options)
    }
}

AppTools.prototype = {

    // Parse out command
    dispatch: function(argv) {
        var action = (argv._ && argv._.length > 0) ? argv._.shift() : '';
        this.action = action.toLowerCase();

        switch(action){
            case '':
                this.helpers.fatal('No command was found.');
                break;
            case 'new-app':             // done
                this.newApp(argv);
                break;
            // QUESTION: rename to publish-app-remote?
                // should all cloud commands have remote in them?!?!
            // if app already exists in cloud, it will update it
            case 'publish-app':         // done
                this.publishApp(argv);
                break;
            // NOTE: this might not be available to users
//            case 'remove-app-remote':
//                this.removeAppRemote(argv);
//                break;
            case 'add-profile':         // done
                this.addProfile(argv);
                break;
            case 'remove-profile':      // done
                this.removeProfile(argv);
                break;
            case 'list-profiles':       // done
                this.listProfiles(argv);
                break;
            case 'list-apps':           // done => need to fix output
                this.listApps(argv);
                break;
            case 'create-app-metadata': // todo
                // TODO
                    // this is for apps not being hosted in cloud
                    // NOTE: need to add custom location (url) where app resides
                break;
            default:
                this.helpers.fatal('Unknown Command: ' + action + '.');
        }
    },

    // Commands
    newApp: function(argv) {    // Args: -t, --title, --description

        // Parse inputs
        var name = argv._.shift(),
            type = argv.t || 'custom';
        if (!name) {
            this.helpers.fatal('Missing project name.');
        }
        var pathToAppFolder = path.join(process.cwd(), 'apps', name);

        this.demandProjectOrExit();

        this.helpers.checkIfAppExists(pathToAppFolder);

        // Prompt for info
        var self = this;
        prompt.message = 'app-tools';
        prompt.override = argv;
        prompt.start();
        prompt.get([
            {
                name: "title",
                description: "display name of app (required)".cyan,
                required: true
            },
            {
                name: "description",
                description: "description of app (optional)".cyan,
                required: false
            }
        ], function(err, result) {

            if (err) {
                self.helpers.fatal('\nAborted app creation.');
            }

            self.helpers.createAppDirectory(pathToAppFolder);

            // Generate skeleton project
            type = type.toLowerCase();
            switch (type) {
                case 'custom':
                    self.helpers.createCustomApp(name, result.title, result.description, pathToAppFolder);
                    break;
                case 'app-generator':
                    // TODO
//                    self.helpers.createAppGenerator(name, result.title, result.description, pathToAppFolder);
                    break;
                case 'angular':
                    // TODO
                    self.helpers.createAngularSeedApp(name, result.title, result.description, pathToAppFolder);
                    break;
                default:
                    // in case user spells it wrong...
                    self.fatal('Unknown project type. Please specify either custom, angular, or app-generator.');
            }

        });

    },

    // ASSUMPTION: you have to be in the app directory to publish!
    publishApp: function(argv) {    // Possible Args: -e -n -u -p --dir (optional -> default to /www)
        // ENDPOINT: PUT or POST to /apps

        // Parse inputs
        var publishDirectory = argv.dir || 'www';

        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);

        // Validate
        this.helpers.checkFileOrPath(pathToAppFolder, publishDirectory);   // Probably don't need...
        this.helpers.checkFileOrPath(pathToAppFolder, publishDirectory, 'index.html');
        this.helpers.checkFileOrPath(pathToAppFolder, publishDirectory, 'package.json');
        this.helpers.checkFileOrPath(pathToAppFolder, publishDirectory, 'LICENSE');

        // Create app instance
        // NOTE: for this command only we will use the package.json from the publish directory
        this.app = this.helpers.loadPackageSync(path.join(pathToAppFolder, publishDirectory));
        this.controllers.app = AppController(this.app);

        // Make sure a project exists so can have an env
        this.demandProjectOrExit();

        // Validate credentials
        var env = this.helpers.createEnv(this.project, argv);
        var namespace = env.getNamespace(),
            username = env.getUser(),
            password = env.getPassword();

        if (!namespace || !username || !password) {
            this.helpers.fatal('Please provide an environment or a namespace, username, and password.');
        }

        // Make sure all the required attributes are present before publishing
        if (!this.controllers.app.validateApp()) {
            this.helpers.fatal('Missing required fields in package.json. Note: both name and version are required.');

            // TODO: prompt for missing required fields?
        }

        // Save overwrites to package.json
        this.app.save();

        // Create Zip
        var self = this;

        this.helpers.createAppZip(this.app.getName(), pathToAppFolder, publishDirectory)
            .then(function(obj) {
                var zipName = obj.zipName,
                    size    = obj.size;

                // Publish App
                return self.helpers.sendApp(zipName, size, pathToAppFolder, namespace, username, password, self.app);

            })
            .then(function() {
                // QUESTION: put messaging here?
            });

    },

    removeAppRemote: function(argv) {

        // ENDPOINT: DELETE /apps/<app name>

        // TODO
        var pathToAppFolder = this.helpers.getPathToAppFolder(this.cwd);

        // Create app instance
        this.app = this.helpers.loadPackageSync(pathToAppFolder);
        this.controllers.app = AppController(this.app);
        var name = this.app.getName();

        // Make sure a project exists so can have an env
        this.demandProjectOrExit();

        // Validate credentials
        var credentials = this.helpers.validateCredentials(this.project, argv);

        // Make sure all the required attributes are present before publishing
        if (!this.controllers.app.validateApp()) {
            this.helpers.fatal('Missing required fields in package.json. Note: both name and version are required.');

            // TODO: prompt for missing required fields?
        }

        var self = this;

        this.helpers.removeAppRemote(credentials, name)
            .then(function() {

                console.log('Successfully removed ' + name + ' from the server.');

            }, function(err) {

                self.helpers.fatal('Failed to remove ' + name + ' from the server.');

            });
    },

    addProfile: function(argv) {

        // ENDPOINT: POST /apps/<app name>/profiles/<profile name (KFC.user)>

        this.profileCurrier('add')(argv);

    },

    removeProfile: function(argv) {

        // ENDPOINT: DELETE /apps/<app name>/<profile name>

        this.profileCurrier('remove')(argv);
    },

    listProfiles: function(argv) {

        // ENDPOINT: GET /apps/<app name>

        this.profileCurrier('list')(argv);
    },

    profileCurrier: function(type) {

        var self = this,
            successMessage = '',
            failMessage = '',
            fn;

        switch(type) {
            case 'add':
                fn = this.helpers.addProfile;
                successMessage = 'Successfully added profile';
                failMessage = 'Failed to add Profile.';
                break;
            case 'remove':
                fn = this.helpers.removeProfile;
                successMessage = 'Successfully removed profile';
                failMessage = 'Failed to remove Profile.';
                break;
            case 'list':
                fn = this.helpers.listProfiles;
                failMessage = 'Failed to get profiles.';
                break;
        }

        return function(argv) {

            var profileName;

            if (type === 'remove' || type === 'add') {
                profileName = argv._.shift();
                if (!profileName) {
                    self.helpers.fatal('Missing profile name. Please specify one.');
                }
            }

            self.demandProjectOrExit();

            // Validate credentials
            var credentials = self.helpers.validateCredentials(self.project, argv);

            var pathToAppFolder = self.helpers.getPathToAppFolder(self.cwd);
            self.app = self.helpers.loadPackageSync(pathToAppFolder);

            fn(self.app.getName(), credentials, profileName)
                .then(function(listOfProfiles) {

                    if (listOfProfiles) {
                        console.log('\nProfiles of ' + self.app.getName() + ':');
                        console.log(listOfProfiles + '\n');
                    } else {
                        console.log(successMessage);
                    }

                }, function(err) {
                    console.log(err);
                    self.helpers.fatal(failMessage + ' Please try again.');
                });
        }
    },

    listApps: function(argv) {

        // ENDPOINT: Get /apps

        var credentials = this.helpers.validateCredentials(this.project, argv),
            self = this;

        this.helpers.listApps(credentials)
            .then(function(apps) {
                console.log(apps);
            }, function() {
                self.helpers.fatal('Failed to get apps. Please try again.');
            });
    },

    // NOTE: these are repeats... abstract this out for all the commands?
    demandProjectOrExit: function(){
        if(!this.project){
            this.helpers.fatal('Unknown project. Please init a new project in this directory or move to an active project.');
        }
    }
};

// MAIN
new AppTools(argv).dispatch(argv);