module.exports = function(options){
    'use strict';

    options = options || {};


    // TODO: put these in the methods that require them!
    // TODO: promises?

    var fs       = require('fs-extra');
    var path     = require('path');
    var findup   = require('findup');
    var FormData = require('form-data');
    var rimraf   = require('rimraf');
    var async    = require('async');
    var Q        = require('q');
    var http     = require('http');

    var config        = require('./config');
    var _             = require('./utils');
    var Project       = require('./project');
    var Environment   = require('./environment');
    var App           = require('./app');

    var PROJECT_FILE  = 'rbc-project.json';
    var PACKAGE_FILE  = 'package.json';

    var NullProject = {
        getDefaultEnvironment: function(){return null;}
        , hasEnvironment     : function(){ return false;}
    };

    function fatal(msg, code) {
        code = code || 5;
        console.log(msg.red);
        process.exit(code);
    }

    function loadProjectSync(){
        var project = new Project();
        try{
            var home = findup.sync(process.cwd(), PROJECT_FILE);
            try{
                if(options.verbose) console.log('Binding to project ' + home)
                process.chdir(home);
                var file       = path.join(home, PROJECT_FILE);
                var spec       = require(file);
                var project    = new Project(spec);
                if(options.verbose) console.log('Project Bound');
                if(options.debug)   console.log('Project', project);
            } catch(err){
                if(options.verbose) console.error('Unable to bind to ' + PROJECT_FILE +  ' in ' + home);
                if(options.debug)   console.error(err);
            }
        } catch(e) {
            if(options.verbose) console.error('Unable to open ' + PROJECT_FILE +  ' in ' + home);
            if(options.debug)   console.error(e);
        }
        return project;
    }

    function loadPackageSync(appHome) {

        try{
            process.chdir(appHome); // QUESTION: do we need to do this?
            var file       = path.join(appHome, PACKAGE_FILE);
            var attr       = require(file);
            var app        = new App(attr);
        } catch(err){
            fatal('Unable to bind to ' + PACKAGE_FILE +  ' in ' + appHome);
            if(options.debug)   console.error(err);
        }

        return app;
    }

    /**
     * Create
     * @param argOpts
     * @returns {{}}
     */
    function createEnvironmentPatch(args){

        args    = args    || {}

        var patch = {}

        //Create a diff
        if (args.uri)       patch.uri       = args.uri;
        if (args.username)  patch.username  = args.username;
        if (args.password)  patch.password  = args.password;
        if (args.namespace) patch.namespace = args.namespace;
        if (args.agent)     patch.agent     = args.agent;

        return patch;
    }

    function lookupEnvironment(project, args){

        project = project || NullProject;
        args    = args    || {}

        var name    = args.environment || project.getDefaultEnvironment();
        var env     = null;

        if(name == null || !project.hasEnvironment(name)){
            if(options.verbose) console.log('Unable to locate an environment');
            env = null;
        } else {
            if(options.verbose) console.log("Using environment " + name);
            env = project
                .getEnvironment(name)
                .patch(createEnvironmentPatch(args))
        }

        if(options.debug) console.log('Env', env);

        return env;
    }

    //add to the project file
    function createEnvironment(project, args){

        project = project || NullProject;
        args    = args    || {}

        var name = args.environment || project.getDefaultEnvironment();
        var env  = null;

        if(name == null || !project.hasEnvironment(name)){
            if(options.verbose) console.log('Using anonymous environment');
            env = new Environment('_$ANONYMOUS$_', {})
        } else {
            if(options.verbose) console.log("Using environment " + name);
            env = project.getEnvironment(name)
        }

        //Patching the environment
        env.patch(createEnvironmentPatch(args))

        if(options.debug) console.log('Env', env);

        return env;
    }


    // TODO: change to promises; fatal should be moved to app-tools
    // ASSUMPTION: publishDirectory is just the name of the dir from the root of the app folder
        // publishDirectory must be located in the root of the app folder!
        // if no publishDirectory => make the assumption it will be in www
        // ENCOURAGE: having production version of the code for performance
    function checkFileOrPath(appPath, publishDirectory, file) {

        file = file || '';

        var publishFolderPath = path.join(appPath, publishDirectory, file);

        try {

            return fs.openSync(publishFolderPath, 'r');

        } catch (e) {
            if (file) {
                fatal(file + ' is not located at the root of ' + publishDirectory + '/');
            } else {
                fatal('Failed to locate ' + publishFolderPath);
            }

            return undefined;
        }
    }

    function createAppZip(name, appPath, publishDirectory) {

        var zipName = name + '-' + new Date().getTime() + '.zip';
        var q = Q.defer();

        var archiver = require('archiver');
        var archive  = archiver('zip');

        var output = fs.createWriteStream(path.join(appPath, zipName));

        output.on('close', function() {
            console.log(zipName + ' has been written to ' + appPath + '/');     // TODO: this should move to app-tools
            q.resolve({size: archive.pointer(), zipName: zipName});
        });

        output.on('error', function(err) {
            fatal('Failed to create zip file');                                 // TODO: this should move to app-tools
            q.reject(); // not sure if need
        });

        archive.pipe(output);

        archive.bulk([
            {expand: true, cwd: path.join(appPath, publishDirectory), src: ['**']}
        ]);

        archive.finalize();

        return q.promise;
    }

    // TODO: refactor - make this cleaner, more organized!
    function sendApp(zipName, size, appPath, namespace, username, password, app) {

        // Dependencies
        var request = require('request');

        var zipPath = path.join(appPath, zipName);

        var zipNameSplit = zipName.split('-');
        zipNameSplit.pop();     // Take out the last segment (date that was tacked on in createAppZip)
        var name = zipNameSplit.join('-');

        var buffer = fs.createReadStream(zipPath);
        var q = Q.defer();

        var urlPath = 'apps';
        var form = new FormData();

        form.append('upload', buffer, {
            filename: name + '.zip',
            contentType: 'application/zip',
            knownLength: size
        });
        form.append('name', app.getName());
        form.append('title', app.getTitle());
        form.append('version', app.getVersion());
        form.append('description', app.getDescription());

        // TODO: implement progress bar for this
            // like int fetchArchive

        console.log('Submitting...');

        // Check if PUT or POST
        var url = 'https://' + path.join(config.SERVER_ENDPOINT, urlPath, name);

        var self = this;

        request({
            method: 'GET',
            uri: url,
            auth: {
                user: username,
                password: password
            }
        }, function(err, response, body) {

            if (err) {
                self.fatal('Failed to publish app. Aborting publish.');
                q.reject(); // TODO: move fatal out to app-tools
            }

            var type;
            response.statusCode !== 200 ? type = 'POST' : type = 'PUT';

            form.submit({
                    /*
                    NOTE:
                        config.SERVER_ENDPOINT must be in the following format:
                            'api.rbcplatform.com'
                        whereas
                            'https://api.rbcplatform.com/'
                        will result to:
                            { [Error: getaddrinfo ENOTFOUND] code: 'ENOTFOUND', errno: 'ENOTFOUND', syscall: 'getaddrinfo' }

                    PATH: must also be /somePathName whereas it was somePathName
                    */
                    host: config.SERVER_ENDPOINT,
                    path: '/' + urlPath,
                    auth: username + ':' + password,
                    /*
                    PROTOCOL REQUIRED ELSE FORM-DATA WILL REQUEST IT AS HTTP INSTEAD OF HTTPS
                    FormData.prototype.submit = function (params, cb) {
                        ...
                          if (params.protocol == 'https:') {
                            request = https.request(options);
                          } else {
                            request = http.request(options);
                          }
                    }
                    */
                    protocol: 'https:',
                    method: type
                }
                , function(err, res) {

                    // Clean up
                    cleanUpPublish(zipPath);    // TODO: this is async => plan for this with promises...

                    if (err || (type === 'PUT' && res.statusCode !== 204) || (type === 'POST' && res.statusCode !== 201)) {
                        console.log(err);
                        fatal('Failed to publish app to ' + config.SERVER_ENDPOINT + '/' + urlPath + '. Aborting publish.');    // TODO: reject w/ this
                        q.reject();
                    }

                    console.log('App published to ' + config.SERVER_ENDPOINT + '/' + urlPath);  // TODO: resolve with this msg

                    res.resume();
                    q.resolve();

                });

        });

        return q.promise;

    }

    // TODO: this should be a promise
    function createCustomApp(name, title, description, pathToAppFolder) {

        fs.copy(path.join(__dirname, 'custom-templates'), pathToAppFolder, function(err) {
            if (err) {

                console.log(err);           // TODO: make debugging flag for this!

                cleanUpNewApp(pathToAppFolder)
                    .then(function() {
                        // TODO: reject w/ msg
                        fatal('Failed to write files to ' + pathToAppFolder + ' Aborted app creation.');
                    });
            }

            writeToPackageJson(name, title, description, pathToAppFolder, path.join(pathToAppFolder, 'package.json'))
                .then(function() {
                    // TODO: resolve with msg
                    console.log('Files have been written to ' + pathToAppFolder);
                })
                .fail(function() {

                    // TODO: resolve msg
                    console.log('Failed to update package.json with name, title, and description. Please update it manually.');
//                    fatal('Failed to update package.json with name, title, and description. Please update it manually.');

                });

        });

    }

    // TODO: should return a promise!
        // helpers should start and end in app-tools.js not in helpers.js...
    // TODO: helpers should have fatal, info, and debug! (user view.js?!?!?!)
    function createAngularSeedApp(name, title, description, pathToAppFolder) {

        fetchArchive(pathToAppFolder, config.ANGULAR_ENDPOINT)
            .then(function() {
                return writeToPackageJson(name, title, description, pathToAppFolder, path.join(pathToAppFolder, 'package.json'));
            })
            .then(function() {
                // DONE
            })
            .fail(function(errMsg) {
                cleanUpNewApp(pathToAppFolder);
                fatal(errMsg);
            });

    }

    // HELPER

    // NOTE: used ionic-cli's fetchArchive function
    function fetchArchive(targetPath, archiveUrl) {

        var os = require('os');
        var q = Q.defer();
        var unzip = require('unzip2');

        console.log('\nDownloading:', archiveUrl);

        var tmpFolder = os.tmpdir();
        var tempZipFilePath = path.join(tmpFolder, 'angular-seed-' + new Date().getTime() + '.zip');

        // HELPER
        var unzipRepo = function(fileName) {

            var readStream = fs.createReadStream(fileName);
            readStream.on('error', function(err) {
                console.log( ('unzipRepo readStream: ' + err).red );
                q.reject(err);
            });

            readStream.on('open', function() {
                readStream.pipe(unzip.Parse())
                    .on('entry', function(entry) {
                        var fileName = entry.path.split('/').slice(1).join('/');

                        if (fileName === "") {
                            entry.autodrain();
                        } else if (entry.type === 'Directory') {
                            fs.ensureDir(path.join(targetPath, fileName));
                        } else {    // file
                            entry.pipe(fs.createWriteStream(path.join(targetPath, fileName)));
                        }
                    })
                    .on('close', function() {
                        console.log('Successfully wrote app files to ' + targetPath);
                        q.resolve();
                    })
                    .on('error', function() {
                        // TODO: add clean up new app to all fatals!
                        q.reject('Failed to write app files to ' + targetPath);
                    });
            });

        };

        // MAIN
        var proxy = process.env.PROXY || process.env.http_proxy || null;
        var request = require('request');
        request({ url: archiveUrl, rejectUnauthorized: false, encoding: null, proxy: proxy }, function(err, res, body) {
            if(err) {
                console.error('Error fetching:'.red, archiveUrl, err);
                q.reject(err);
                return;
            }
            if(!res) {
                console.error('Invalid response:'.red, archiveUrl);
                q.reject('Unable to fetch response: ' + archiveUrl);
                return;
            }
            if(res.statusCode !== 200) {
                if(res.statusCode === 404 || res.statusCode === 406) {
                    console.error('Not found:'.red, archiveUrl, '(' + res.statusCode + ')');
                    console.error('Please verify the url and try again.'.red);
                } else {
                    console.error('Invalid response status:'.red, archiveUrl, '(' + res.statusCode + ')');
                }
                q.reject(res);
                return;
            }
            try {
                fs.writeFileSync(tempZipFilePath, body);
                unzipRepo(tempZipFilePath);
            } catch(e) {
                console.log('fetchArchive request write: ' + e);
                q.reject(e);
            }
        }).on('response', function(res){
                var ProgressBar = require('progress');
                var bar = new ProgressBar('[:bar]  :percent  :etas', {
                    complete: '=',
                    incomplete: ' ',
                    width: 30,
                    total: parseInt(res.headers['content-length'], 10)
                });

                res.on('data', function (chunk) {
                    try {
                        bar.tick(chunk.length);
                    } catch(e){}
                });
            });

        return q.promise;
    }

    function writeToPackageJson(name, title, description, pathToAppFolder, srcPackageJson) {

        var q = Q.defer();

        fs.readFile(srcPackageJson, function(err, packageObj) {

            if (err) {
                q.reject('Failed to read package.json template. Please try again.\n' + err);
            }

            var compiledJson = _.template(packageObj, {
                name: name,
                title: title,
                description: description
            });

            // write file
            fs.writeFile(path.join(pathToAppFolder, '/package.json'), compiledJson, function(err) {
                if (err) {
                    q.reject('Failed to create Custom Skeleton App. Please try again.');
                }
                console.log('package.json has been written to ' + pathToAppFolder);
                q.resolve();    // TODO: resolve w/ msg
            });
        });

        return q.promise;
    }

    // TODO: test
    function cleanUpNewApp(pathToAppFolder) {

        var q = Q.defer();
        // Delete app directory
        console.log(pathToAppFolder);
        rimraf(pathToAppFolder, function(err) {
            if (err) {
                fatal('Could not delete ' + pathToAppFolder + '. Please manually delete it.');
            }

            console.log('Finished cleanup.');

            q.resolve();
        });

        return q.promise;

    }

    // TODO: test
    function cleanUpPublish(zipPath) {
        fs.unlink(zipPath, function(err) {
            if (err) {
                fatal('Could not delete app zip. Please manually delete it.');
            }

            console.log('Finished cleanup.');
        });
    }

    function getPathToAppFolder(cwd) {

        // Find name of app
        var split = cwd.split('/'),
            index = split.indexOf('apps'),
            lastIndex = split.length-1;

        if (index > 0 &&                // apps exist in path
            lastIndex - index > 0) {    // not in apps folder
            var name = split[index+1];  // right after apps
        } else {
            fatal('Not in an app directory! Please cd into your app directory and try again.');
        }

        return path.join(process.cwd(), '/apps/', name);

    }

    function checkIfAppExists(pathToAppFolder) {
        if (fs.existsSync(pathToAppFolder)) {
            fatal('App name already exists! Please choose a different name.');
        }

        return true;
    }

    function createAppDirectory(pathToAppFolder) {
        // Create project directory
        fs.mkdirpSync(pathToAppFolder);   // TODO: make this async
    }

    function profileHelperCurrier(type) {

        // add namespace if profileName doesn't specify one?

        return function(name, credentials, profileName) {

            // dependencies
            var prettyJSON = require('prettyjson');

            var request = require('request'),
                q = Q.defer(),
                namespace = credentials.namespace,
                username = credentials.username,
                password = credentials.password,
                url;

            if (type === 'get') {
                url = 'https://' + path.join(config.SERVER_ENDPOINT, 'apps', name);
            } else {
                url = 'https://' + path.join(config.SERVER_ENDPOINT, 'apps', name, 'profiles', profileName);
            }

            request({
                method: type,
                uri: url,
                auth: {
                    user: username,
                    password: password
                }
            }, function(err, response, body) {

                if (err || ( ((type === 'post' || type === 'delete') && response.statusCode !== 204)) || (type === 'get' && response.statusCode !== 200)) {
                    q.reject(err);
                } else {

                    // REMOVE
//                    console.log(JSON.parse(JSON.parse(body)));
                    // REMOVE

                    if (type === 'get') {
                        // TODO: make pretty

                        q.resolve(prettyJSON.render(JSON.parse(body).profiles));
//                        q.resolve(prettyJSON.render(JSON.parse(JSON.parse(body)).profiles));
                    } else {
                        q.resolve();
                    }

                }

            });

            return q.promise;
        };
    }

    function validateCredentials(project, argv) {
        var env = createEnvironment(project, argv);
        var namespace = env.getNamespace(),
            username = env.getUser(),
            password = env.getPassword();

        if (!namespace || !username || !password) {
            this.helpers.fatal('Please provide an environment or a namespace, username, and password.');
        }

        return {
            namespace: namespace,
            username: username,
            password: password
        };
    }

    function listApps(credentials) {

        // Dependencies
        var prettyJSON = require('prettyjson');

        var request = require('request');
        var namespace = credentials.namespace,
            username = credentials.username,
            password = credentials.password,
            q = Q.defer();

        var url = 'https://' + path.join(config.SERVER_ENDPOINT, 'apps');

        request({
            method: 'GET',
            uri: url,
            auth: {
                user: username,
                password: password
            }
        }, function(err, response, body) {

            if (err || response.statusCode !== 200) {
                q.reject();
            } else {
                q.resolve(process(JSON.parse(body)));
            }

            function process(body) {
                var processedBody = '\n';

                _.each(body, function(app) {

//                    app = JSON.parse(app);

//                    delete app.rbc_modified_at;
                    delete app.rbc_type_version;
                    delete app.rbc_type_version;
                    delete app.rbc_modified_by;
                    delete app.rbc_created_by;
//                    delete app.rbc_created_at;
                    delete app.rbc_created_namespace;

                    processedBody = processedBody + prettyJSON.render(app) + '\n\n';

                });

                return processedBody;
            }

        });

        return q.promise;

    }

    function removeAppRemote(credentials, name) {
        // Dependencies
        var request = require('request');

        var namespace = credentials.namespace,
            username = credentials.username,
            password = credentials.password,
            q = Q.defer();

        var url = 'https://' + path.join(config.SERVER_ENDPOINT, 'apps', name);
        console.log(url);

        request({
            method: 'DELETE',
            uri: url,
            auth: {
                user: username,
                password: password
            }
        }, function(err, response, body) {
            console.log(err);
            console.log(response.statusCode);
            if (err || response.statusCode !== 204) {
                q.reject();
            } else {
                q.resolve();
            }
        });

        return q.promise;
    }

    return {
          fatal           : fatal
        , loadProjectSync : loadProjectSync
        , createEnv       : createEnvironment
        , lookupEnv       : lookupEnvironment
        , createEnvPatch  : createEnvironmentPatch

        // package.json
        , loadPackageSync : loadPackageSync

        // New App
        , createAppDirectory: createAppDirectory

        // Skeleton Apps
        , createCustomApp : createCustomApp
        , createAngularSeedApp : createAngularSeedApp

        // Checks
        , checkFileOrPath : checkFileOrPath

        // App Zip
        , createAppZip    : createAppZip

        // Associate App
        , addProfile      : profileHelperCurrier('post')
        , removeProfile   : profileHelperCurrier('delete')
        , listProfiles    : profileHelperCurrier('get')

        // Publish App
        , getPathToAppFolder: getPathToAppFolder
        , sendApp         : sendApp

        // Remove App
        , removeAppRemote: removeAppRemote

        , listApps         : listApps

        // Helpers
        , validateCredentials: validateCredentials
        , checkIfAppExists: checkIfAppExists

        // Files
        , PROJECT_FILE    : PROJECT_FILE
        , PACKAGE_FILE    : PACKAGE_FILE
    }
};